@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2011  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2012  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -74,6 +74,14 @@ static GHashTable *network_table;
 
 static int dhcpv6_request(struct connman_dhcpv6 *dhcp, gboolean add_addresses);
 
+static void clear_timer(struct connman_dhcpv6 *dhcp)
+{
+	if (dhcp->timeout > 0) {
+		g_source_remove(dhcp->timeout);
+		dhcp->timeout = 0;
+	}
+}
+
 static inline float get_random()
 {
 	return (rand() % 200 - 100) / 1000.0;
@@ -335,19 +343,27 @@ static int dhcpv6_info_request(struct connman_dhcpv6 *dhcp)
 	index = connman_network_get_index(dhcp->network);
 
 	dhcp_client = g_dhcp_client_new(G_DHCP_IPV6, index, &error);
-	if (error != G_DHCP_CLIENT_ERROR_NONE)
+	if (error != G_DHCP_CLIENT_ERROR_NONE) {
+		clear_timer(dhcp);
 		return -EINVAL;
+	}
 
 	if (getenv("CONNMAN_DHCPV6_DEBUG"))
 		g_dhcp_client_set_debug(dhcp_client, dhcpv6_debug, "DHCPv6");
 
 	service = __connman_service_lookup_from_network(dhcp->network);
-	if (service == NULL)
+	if (service == NULL) {
+		clear_timer(dhcp);
+		g_dhcp_client_unref(dhcp_client);
 		return -EINVAL;
+	}
 
 	ret = set_duid(service, dhcp->network, dhcp_client, index);
-	if (ret < 0)
+	if (ret < 0) {
+		clear_timer(dhcp);
+		g_dhcp_client_unref(dhcp_client);
 		return ret;
+	}
 
 	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_CLIENTID);
 	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_DNS_SERVERS);
@@ -808,10 +824,7 @@ int __connman_dhcpv6_start_renew(struct connman_network *network,
 
 	DBG("network %p dhcp %p", network, dhcp);
 
-	if (dhcp->timeout > 0) {
-		g_source_remove(dhcp->timeout);
-		dhcp->timeout = 0;
-	}
+	clear_timer(dhcp);
 
 	g_dhcpv6_client_get_timeouts(dhcp->dhcp_client, &T1, &T2,
 				&last_renew, &last_rebind, &expired);
@@ -882,10 +895,7 @@ int __connman_dhcpv6_start_release(struct connman_network *network,
 	if (dhcp->stateless == TRUE)
 		return -EINVAL;
 
-	if (dhcp->timeout > 0) {
-		g_source_remove(dhcp->timeout);
-		dhcp->timeout = 0;
-	}
+	clear_timer(dhcp);
 
 	dhcp_client = dhcp->dhcp_client;
 	if (dhcp_client == NULL) {
@@ -929,10 +939,7 @@ static int dhcpv6_release(struct connman_dhcpv6 *dhcp)
 {
 	DBG("dhcp %p", dhcp);
 
-	if (dhcp->timeout > 0) {
-		g_source_remove(dhcp->timeout);
-		dhcp->timeout = 0;
-	}
+	clear_timer(dhcp);
 
 	dhcpv6_free(dhcp);
 
@@ -1032,10 +1039,7 @@ static void advertise_cb(GDHCPClient *dhcp_client, gpointer user_data)
 
 	DBG("dhcpv6 advertise msg %p", dhcp);
 
-	if (dhcp->timeout > 0) {
-		g_source_remove(dhcp->timeout);
-		dhcp->timeout = 0;
-	}
+	clear_timer(dhcp);
 
 	if (g_dhcpv6_client_get_status(dhcp_client) != 0) {
 		if (dhcp->callback != NULL)
@@ -1059,10 +1063,7 @@ static void solicitation_cb(GDHCPClient *dhcp_client, gpointer user_data)
 
 	DBG("dhcpv6 solicitation msg %p", dhcp);
 
-	if (dhcp->timeout > 0) {
-		g_source_remove(dhcp->timeout);
-		dhcp->timeout = 0;
-	}
+	clear_timer(dhcp);
 
 	set_addresses(dhcp_client, dhcp);
 }
@@ -1095,19 +1096,27 @@ static int dhcpv6_solicitation(struct connman_dhcpv6 *dhcp)
 	index = connman_network_get_index(dhcp->network);
 
 	dhcp_client = g_dhcp_client_new(G_DHCP_IPV6, index, &error);
-	if (error != G_DHCP_CLIENT_ERROR_NONE)
+	if (error != G_DHCP_CLIENT_ERROR_NONE) {
+		clear_timer(dhcp);
 		return -EINVAL;
+	}
 
 	if (getenv("CONNMAN_DHCPV6_DEBUG"))
 		g_dhcp_client_set_debug(dhcp_client, dhcpv6_debug, "DHCPv6");
 
 	service = __connman_service_lookup_from_network(dhcp->network);
-	if (service == NULL)
+	if (service == NULL) {
+		clear_timer(dhcp);
+		g_dhcp_client_unref(dhcp_client);
 		return -EINVAL;
+	}
 
 	ret = set_duid(service, dhcp->network, dhcp_client, index);
-	if (ret < 0)
+	if (ret < 0) {
+		clear_timer(dhcp);
+		g_dhcp_client_unref(dhcp_client);
 		return ret;
+	}
 
 	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_CLIENTID);
 	g_dhcp_client_set_request(dhcp_client, G_DHCPV6_RAPID_COMMIT);
